import numpy as np
from numpy import einsum
from .MaterialBase import Material
from Florence.Tensor import trace, Voigt

class TranservselyIsotropicHyperElastic(Material):
    """A compressible transervely isotropic model with the isotropic part being Mooney-Rivlin
        The energy is given by:

            W(C) =  gamma * ( alpha*(C:I) + beta*(G:I) ) +
                    eta*(1-alpha)*( (N C N)**2 + N G N) - ut*J + lambda/2*(J-1)**2

            ut = 2.*gamma*(alpha+2.0*beta) + 2.*(1. - gamma)*eta  # for the stress to be
                zero at the origin

        the parameter "gamma" controls the amount of anisotropy and the vector N(ndim,1) is
        the direction of anisotropy

    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(TranservselyIsotropicHyperElastic, self).__init__(mtype, ndim, **kwargs)
        self.ndim = ndim
        self.nvar = self.ndim

        self.is_transversely_isotropic = True
        self.is_nonisotropic = True
        self.energy_type = "internal_energy"
        self.nature = "linear"
        self.fields = "mechanics"

        if self.ndim==3:
            self.H_VoigtSize = 6
        elif self.ndim==2:
            self.H_VoigtSize = 3

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = False

        self.gamma = 0.5



    def Hessian(self,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):

        # Get material constants (5 in this case)
        E = self.E
        E_A = self.E_A
        v = self.nu

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        F = StrainTensors['F'][gcounter]
        H = J*np.linalg.inv(F).T
        N = self.anisotropic_orientations[elem][:,None]
        FN = np.dot(F,N)[:,0]
        HN = np.dot(H,N)[:,0]
        innerFN = einsum('i,i',FN,FN)
        innerHN = einsum('i,i',HN,HN)
        outerHN = einsum('i,j',HN,HN)

        gamma = self.gamma

        lamb = -(E_A*E*v)/(2.*E*v**2 + E_A*v - E_A)
        ut = (E**2*v**2 + E_A*E*v**2 + E_A*E*v - E_A*E)/(2*(v + 1)*(2*E*v**2 + E_A*v - E_A))
        beta = 0.
        eta_1 = (E_A**2*v**2 - E_A**2 - 2*E_A*E*v**2 + E_A*E + E**2*v**2)/(4*(gamma - 1)*(v + 1)*(2*E*v**2 + E_A*v - E_A))
        eta_2 = -(E_A**2*v - E_A**2 + E_A*E - E_A*E*v)/(4*(gamma - 1)*(2*E*v**2 + E_A*v - E_A))
        alpha = ut - 4*gamma*beta - 2*(1-gamma)*eta_1 - 2*(1-gamma)*eta_2
        alpha = alpha/2./gamma
        eta = [eta_1,eta_2]


        H_Voigt = 2.*gamma*beta/J* ( 2.0*einsum('ij,kl',b,b) - einsum('ik,jl',b,b) - einsum('il,jk',b,b) ) - \
                (- lamb*(2.*J-1.) ) *einsum('ij,kl',I,I) + \
                (ut - lamb*(J-1.) ) * ( einsum('ik,jl',I,I) + einsum('il,jk',I,I) )



        for m in range(2,4):
            H_Voigt += self.TransverseHessianNCN(StrainTensors,m,eta[m-2],gamma,FN,innerFN,elem,gcounter)

        H_Voigt += self.TransverseHessianNGN(StrainTensors,1.,eta_1,gamma,HN,innerHN,elem,gcounter)
        H_Voigt += self.TransverseHessianNGN(StrainTensors,1.,eta_2,gamma,HN,innerHN,elem,gcounter)

        H_Voigt = Voigt(H_Voigt ,1)

        self.H_VoigtSize = H_Voigt.shape[0]

        return H_Voigt


    def TransverseHessianNCN(self,StrainTensors,m,eta,gamma,FN,innerFN,elem,gcounter):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]

        H_VoigtNCN = 4.*(1-gamma)*eta/J *(m-1)*(innerFN)**(m-2)*einsum('i,j,k,l',FN,FN,FN,FN)

        return H_VoigtNCN

    def TransverseHessianNGN(self,StrainTensors,n,eta,gamma,HN,innerHN,elem,gcounter):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]

        H_VoigtNGN = 4.*(1-gamma)*eta/J * ( n*(innerHN)**n * einsum('ij,kl',I,I) - \
                0.5*(innerHN)**n * ( einsum('ik,jl',I,I) + einsum('il,jk',I,I) ) - \
                n*(innerHN)**(n-1)* ( einsum('ij,k,l',I,HN,HN) + einsum('i,j,kl',HN,HN,I) ) + \
                (n-1.)*(innerHN)**(n-2)* einsum('i,j,k,l',HN,HN,HN,HN) ) + \
                2.*(1-gamma)*eta/J *(innerHN)**(n-1)* ( einsum('il,j,k',I,HN,HN) + einsum('jl,i,k',I,HN,HN) + \
                einsum('ik,j,l',I,HN,HN) + einsum('jk,i,l',I,HN,HN) )

        return H_VoigtNGN






    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        F = StrainTensors['F'][gcounter]
        H = J*np.linalg.inv(F).T
        N = self.anisotropic_orientations[elem][:,None]
        FN = np.dot(F,N)
        HN = np.dot(H,N)[:,0]
        innerFN = np.dot(FN.T,FN)[0][0]
        innerHN = einsum('i,i',HN,HN)
        outerHN = einsum('i,j',HN,HN)

        E = self.E
        E_A = self.E_A
        v = self.nu

        gamma = self.gamma

        lamb = -(E_A*E*v)/(2.*E*v**2 + E_A*v - E_A)
        ut = (E**2*v**2 + E_A*E*v**2 + E_A*E*v - E_A*E)/(2*(v + 1)*(2*E*v**2 + E_A*v - E_A))
        beta = 0.
        eta_1 = (E_A**2*v**2 - E_A**2 - 2*E_A*E*v**2 + E_A*E + E**2*v**2)/(4*(gamma - 1)*(v + 1)*(2*E*v**2 + E_A*v - E_A))
        eta_2 = -(E_A**2*v - E_A**2 + E_A*E - E_A*E*v)/(4*(gamma - 1)*(2*E*v**2 + E_A*v - E_A))
        alpha = ut - 4*gamma*beta - 2*(1-gamma)*eta_1 - 2*(1-gamma)*eta_2
        alpha = alpha/2./gamma
        eta = [eta_1,eta_2]


        if self.ndim == 3:
            trb = trace(b)
        elif self.ndim == 2:
            trb = trace(b) + 1


        stress = 2.*gamma*alpha/J*b + 2.*gamma*beta/J*(trb*b - np.dot(b,b)) - ut/J*I + lamb*(J-1.)*I

        # n=1
        # m=2
        # stressNCN_1 = self.CauchyStressNCN(StrainTensors,m,eta,gamma,FN,innerFN,elem,gcounter)
        # stressNGN_1 = self.CauchyStressNGN(StrainTensors,n,eta,gamma,innerHN,outerHN,elem,gcounter)
        # stress += stressNCN_1
        # stress += stressNGN_1

        for m in range(2,4):
            stress += self.CauchyStressNCN(StrainTensors,m,eta[m-2],gamma,FN,innerFN,elem,gcounter)

        stress += self.CauchyStressNGN(StrainTensors,1.,eta_1,gamma,innerHN,outerHN,elem,gcounter)
        stress += self.CauchyStressNGN(StrainTensors,1.,eta_2,gamma,innerHN,outerHN,elem,gcounter)

        # print stress
        return stress


    def CauchyStressNCN(self,StrainTensors,m,eta,gamma,FN,innerFN,elem,gcounter):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]

        return 2.*(1.- gamma)*eta/J*(innerFN)**(m-1)*np.dot(FN,FN.T)

    def CauchyStressNGN(self,StrainTensors,n,eta,gamma,innerHN,outerHN,elem,gcounter):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]

        return 2.*(1.- gamma)*eta/J*(innerHN)**(n-1)*(innerHN*I - outerHN)






import numpy as np

# Modal Bases
# import Florence.FunctionSpace.TwoDimensional.Tri.hpModal as Tri
# import Florence.FunctionSpace.ThreeDimensional.Tet.hpModal as Tet
# Nodal Bases
from Florence.FunctionSpace import Line
from Florence.FunctionSpace import Tri
from Florence.FunctionSpace import Tet
from Florence.FunctionSpace import Quad, QuadES
from Florence.FunctionSpace import Hex, HexES

from Florence.QuadratureRules.FeketePointsTri import *
from Florence.QuadratureRules.FeketePointsTet import *
from Florence.QuadratureRules.GaussLobattoPoints import *
from Florence.QuadratureRules.EquallySpacedPoints import *


def GetBases1D(C, Quadrature, info=None, bases_type="nodal", equally_spaced=False, is_flattened=False):

    w = Quadrature.weights
    z = Quadrature.points

    ns=[]; Basis=[]; gBasis = []
    ns = int(C+2)
    Basis = np.zeros((ns,z.shape[0]),dtype=np.float64)
    gBasis = np.zeros((ns,z.shape[0]),dtype=np.float64)

    if not equally_spaced:
        hpBases = Line.LagrangeGaussLobatto
    else:
        hpBases = Line.Lagrange

    for i in range(0,w.shape[0]):
        ndummy, dummy = hpBases(C,z[i])[:2]
        Basis[:,i] = ndummy[:]
        gBasis[:,i] = dummy[:]


    class Domain(object):
        Bases = Basis
        gBasesx = gBasis
        gBasesy = np.zeros(gBasis.shape)
        gBasesz = np.zeros(gBasis.shape)


    return Domain



def GetBases2D(C, Quadrature, info, bases_type="nodal", equally_spaced=False, is_flattened=False):

    w = Quadrature.weights
    z = Quadrature.points

    ns=[]; Basis=[]; gBasisx=[]; gBasisy=[]
    if info=='tri':
        p=C+1
        ns = int((p+1)*(p+2)/2)
        Basis = np.zeros((ns,w.shape[0]),dtype=np.float64)
        gBasisx = np.zeros((ns,w.shape[0]),dtype=np.float64)
        gBasisy = np.zeros((ns,w.shape[0]),dtype=np.float64)
    elif info=='quad':
        ns = int((C+2)**2)
        if is_flattened is False:
            Basis = np.zeros((ns,z.shape[0]*z.shape[0]),dtype=np.float64)
            gBasisx = np.zeros((ns,z.shape[0]*z.shape[0]),dtype=np.float64)
            gBasisy = np.zeros((ns,z.shape[0]*z.shape[0]),dtype=np.float64)
        else:
            Basis = np.zeros((ns,w.shape[0]),dtype=np.float64)
            gBasisx = np.zeros((ns,w.shape[0]),dtype=np.float64)
            gBasisy = np.zeros((ns,w.shape[0]),dtype=np.float64)

    if info == 'quad':
        if not equally_spaced:
            hpBases = Quad.LagrangeGaussLobatto
            GradhpBases = Quad.GradLagrangeGaussLobatto
        else:
            hpBases = QuadES.Lagrange
            GradhpBases = QuadES.GradLagrange

        if is_flattened is False:
            counter = 0
            for i in range(0,z.shape[0]):
                for j in range(0,z.shape[0]):
                    ndummy = hpBases(C,z[i],z[j])
                    Basis[:,counter] = ndummy[:,0]
                    dummy = GradhpBases(C,z[i],z[j])
                    gBasisx[:,counter] = dummy[:,0]
                    gBasisy[:,counter] = dummy[:,1]
                    counter+=1
        else:
            for i in range(0,w.shape[0]):
                ndummy = hpBases(C,z[i,0],z[i,1])
                dummy = GradhpBases(C,z[i,0],z[i,1])
                Basis[:,i] = ndummy[:,0]
                gBasisx[:,i] = dummy[:,0]
                gBasisy[:,i] = dummy[:,1]

    elif info == 'tri':
        hpBases = Tri.hpNodal.hpBases
        for i in range(0,w.shape[0]):
            # Better convergence for curved meshes when Quadrature.optimal!=0
            ndummy, dummy = hpBases(C,z[i,0],z[i,1], Quadrature.optimal, equally_spaced=equally_spaced)
            # ndummy, dummy = Tri.hpBases(C,z[i,0],z[i,1])
            Basis[:,i] = ndummy
            gBasisx[:,i] = dummy[:,0]
            gBasisy[:,i] = dummy[:,1]



    class Domain(object):
        Bases = Basis
        gBasesx = gBasisx
        gBasesy = gBasisy
        gBasesz = np.zeros(gBasisx.shape)


    return Domain



def GetBases3D(C, Quadrature, info, bases_type="nodal", equally_spaced=False, is_flattened=False):

    ndim = 3

    w = Quadrature.weights
    z = Quadrature.points

    ns=[]; Basis=[]; gBasisx=[]; gBasisy=[]; gBasisz=[]
    if info=='hex':
        ns = int((C+2)**ndim)
        if is_flattened is False:
            Basis = np.zeros((ns,(z.shape[0])**ndim),dtype=np.float64)
            gBasisx = np.zeros((ns,(z.shape[0])**ndim),dtype=np.float64)
            gBasisy = np.zeros((ns,(z.shape[0])**ndim),dtype=np.float64)
            gBasisz = np.zeros((ns,(z.shape[0])**ndim),dtype=np.float64)
        else:
            Basis = np.zeros((ns,w.shape[0]),dtype=np.float64)
            gBasisx = np.zeros((ns,w.shape[0]),dtype=np.float64)
            gBasisy = np.zeros((ns,w.shape[0]),dtype=np.float64)
            gBasisz = np.zeros((ns,w.shape[0]),dtype=np.float64)
    elif info=='tet':
        p=C+1
        ns = int((p+1)*(p+2)*(p+3)/6)
        Basis = np.zeros((ns,w.shape[0]),dtype=np.float64)
        gBasisx = np.zeros((ns,w.shape[0]),dtype=np.float64)
        gBasisy = np.zeros((ns,w.shape[0]),dtype=np.float64)
        gBasisz = np.zeros((ns,w.shape[0]),dtype=np.float64)


    if info=='hex':
        if not equally_spaced:
            hpBases = Hex.LagrangeGaussLobatto
            GradhpBases = Hex.GradLagrangeGaussLobatto
        else:
            hpBases = HexES.Lagrange
            GradhpBases = HexES.GradLagrange

        if is_flattened is False:
            counter = 0
            for i in range(w.shape[0]):
                for j in range(w.shape[0]):
                    for k in range(w.shape[0]):
                        ndummy = hpBases(C,z[i],z[j],z[k])
                        dummy = GradhpBases(C,z[i],z[j],z[k])

                        Basis[:,counter] = ndummy[:,0]
                        gBasisx[:,counter] = dummy[:,0]
                        gBasisy[:,counter] = dummy[:,1]
                        gBasisz[:,counter] = dummy[:,2]
                        counter+=1
        else:
            for i in range(0,w.shape[0]):
                ndummy = hpBases(C,z[i,0],z[i,1],z[i,2])
                dummy = GradhpBases(C,z[i,0],z[i,1],z[i,2])
                Basis[:,i] = ndummy[:,0]
                gBasisx[:,i] = dummy[:,0]
                gBasisy[:,i] = dummy[:,1]
                gBasisz[:,i] = dummy[:,2]


    elif info=='tet':
        hpBases = Tet.hpNodal.hpBases
        for i in range(0,w.shape[0]):
            # Better convergence for curved meshes when Quadrature.optimal!=0
            ndummy, dummy = hpBases(C,z[i,0],z[i,1],z[i,2],Quadrature.optimal, equally_spaced=equally_spaced)
            # ndummy, dummy = Tet.hpBases(C,z[i,0],z[i,1],z[i,2])
            Basis[:,i] = ndummy
            gBasisx[:,i] = dummy[:,0]
            gBasisy[:,i] = dummy[:,1]
            gBasisz[:,i] = dummy[:,2]


    class Domain(object):
        """docstring for Domain"""
        Bases = Basis
        gBasesx = gBasisx
        gBasesy = gBasisy
        gBasesz = gBasisz


    return Domain




def GetBasesAtNodes(C, Quadrature, info, bases_type="nodal", equally_spaced=False):

    ns=[]; Basis=[]; gBasisx=[]; gBasisy=[]; gBasisz=[]
    if info == 'hex' or info == "quad":
        w=2
        if info=="quad": ndim = 2
        elif info=="hex": ndim = 3
        ns = int((C+2)**ndim)
        # GET THE BASES AT NODES INSTEAD OF GAUSS POINTS
        Basis = np.zeros((ns,ns))
        gBasisx = np.zeros((ns,ns))
        gBasisy = np.zeros((ns,ns))
        gBasisz = np.zeros((ns,ns))
    elif info =='tet':
        p=C+1
        ns = int((p+1)*(p+2)*(p+3)/6)
        # GET BASES AT NODES INSTEAD OF GAUSS POINTS
        Basis = np.zeros((ns,ns))
        gBasisx = np.zeros((ns,ns))
        gBasisy = np.zeros((ns,ns))
        gBasisz = np.zeros((ns,ns))
    elif info =='tri':
        p=C+1
        ns = int((p+1)*(p+2)/2)
        # GET BASES AT NODES INSTEAD OF GAUSS POINTS
        Basis = np.zeros((ns,ns))
        gBasisx = np.zeros((ns,ns))
        gBasisy = np.zeros((ns,ns))
    elif info == 'line':
        ns = int(C+2)
        # GET THE BASES AT NODES INSTEAD OF GAUSS POINTS
        Basis = np.zeros((ns,ns))
        gBasisx = np.zeros((ns,ns))


    eps=[]
    if info == 'hex':
        counter = 0
        if not equally_spaced:
            eps = GaussLobattoPointsHex(C)
            hpBases = Hex.LagrangeGaussLobatto
            GradhpBases = Hex.GradLagrangeGaussLobatto
        else:
            eps = EquallySpacedPoints(4,C)
            hpBases = HexES.Lagrange
            GradhpBases = HexES.GradLagrange

        for i in range(0,eps.shape[0]):
            ndummy = hpBases(C,eps[i,0],eps[i,1],eps[i,2],arrange=1)
            Basis[:,counter] = ndummy[:,0]
            dummy = GradhpBases(C,eps[i,0],eps[i,1],eps[i,2],arrange=1)
            gBasisx[:,counter] = dummy[:,0]
            gBasisy[:,counter] = dummy[:,1]
            gBasisz[:,counter] = dummy[:,2]
            counter +=1

    elif info == 'quad':
        if not equally_spaced:
            eps = GaussLobattoPointsQuad(C)
            hpBases = Quad.LagrangeGaussLobatto
            GradhpBases = Quad.GradLagrangeGaussLobatto
        else:
            eps = EquallySpacedPoints(3,C)
            hpBases = QuadES.Lagrange
            GradhpBases = QuadES.GradLagrange

        counter = 0
        for i in range(0,eps.shape[0]):
            ndummy = hpBases(C,eps[i,0],eps[i,1],arrange=1)
            Basis[:,counter] = ndummy[:,0]
            dummy = GradhpBases(C,eps[i,0],eps[i,1],arrange=1)
            gBasisx[:,counter] = dummy[:,0]
            gBasisy[:,counter] = dummy[:,1]
            counter+=1

    elif info == 'tet':
        if not equally_spaced:
            eps = FeketePointsTet(C)
        else:
            eps = EquallySpacedPointsTet(C)
        hpBases = Tet.hpNodal.hpBases

        counter = 0
        for i in range(0,eps.shape[0]):
            ndummy, dummy = hpBases(C,eps[i,0],eps[i,1],eps[i,2],1,1, equally_spaced=equally_spaced)
            ndummy = ndummy.reshape(ndummy.shape[0],1)
            Basis[:,counter] = ndummy[:,0]
            gBasisx[:,counter] = dummy[:,0]
            gBasisy[:,counter] = dummy[:,1]
            gBasisz[:,counter] = dummy[:,2]
            counter+=1

    elif info == 'tri':
        if not equally_spaced:
            eps = FeketePointsTri(C)
        else:
            eps = EquallySpacedPointsTri(C)
        hpBases = Tri.hpNodal.hpBases

        for i in range(0,eps.shape[0]):
            ndummy, dummy = hpBases(C,eps[i,0],eps[i,1],1,1, equally_spaced=equally_spaced)
            ndummy = ndummy.reshape(ndummy.shape[0],1)
            Basis[:,i] = ndummy[:,0]
            gBasisx[:,i] = dummy[:,0]
            gBasisy[:,i] = dummy[:,1]

    elif info == 'line':
        if not equally_spaced:
            eps = GaussLobattoQuadrature(C+2)[0]
            hpBases = Line.LagrangeGaussLobatto
        else:
            eps = EquallySpacedPoints(2,C)
            hpBases = Line.Lagrange

        # We probably need node arrangment for lines
        counter = 0
        for i in range(0,eps.shape[0]):
            ndummy = hpBases(C,eps[i,0])
            Basis[:,counter] = ndummy[0][i]
            gBasisx[:,counter] = ndummy[1][i]
            counter+=1



    class Domain(object):
        Bases = Basis
        gBasesx = gBasisx
        gBasesy = gBasisy
        w = np.ones(eps.shape[0])

    if info == "hex" or info == "tet":
        Domain.gBasesz = gBasisz
    elif info == "tri" or info == "quad":
        Domain.gBasesz = np.zeros_like(gBasisx)
    elif info == "line":
        Domain.gBasesy = np.zeros_like(gBasisx)
        Domain.gBasesz = np.zeros_like(gBasisx)

    if info == "hex" or info == "quad":
        Domain.w = np.ones(C+2)

    return Domain



def GetBoundaryBases(C, Quadrature, info, bases_type="nodal", equally_spaced=False, is_flattened=False):

    binfo, ndim = None, None
    if info == "hex":
        binfo = "quad"
        ndim == 3
    elif info == "tet":
        binfo = "tri"
        ndim == 3
    elif info == "quad" or info == "tri":
        binfo = "line"
        ndim == 2

    if Quadrature is None:
        norder = C+2
        from Florence import QuadratureRule
        Quadrature = QuadratureRule(norder=norder, mesh_type=binfo, is_flattened=is_flattened)


    if ndim == 3:
        return GetBases2D(C, Quadrature, info, bases_type=bases_type, equally_spaced=equally_spaced, is_flattened=is_flattened)
    elif ndim == 2:
        return GetBases1D(C, Quadrature, info, bases_type=bases_type, equally_spaced=equally_spaced, is_flattened=is_flattened)






########################################################
# DEPRECATED
# def GetBasesBoundary(C, z, ndim):

#     BasisBoundary = np.zeros(((C+2)**(ndim),(z.shape[0])**(ndim-1),2*ndim))
#     gBasisBoundaryx = np.zeros(((C+2)**(ndim),(z.shape[0])**(ndim-1),2*ndim))
#     gBasisBoundaryy = np.zeros(((C+2)**(ndim),(z.shape[0])**(ndim-1),2*ndim))
#     gBasisBoundaryz = np.zeros(((C+2)**(ndim),(z.shape[0])**(ndim-1),2*ndim))

#     # eps = OneD.LagrangeGaussLobatto(C,0)
#     eps = np.array([-1.,1.,-1.,1.,-1.,1.])


#     for k in range(0,eps.shape[0]):
#         counter = 0
#         for i in range(0,z.shape[0]):
#             for j in range(0,z.shape[0]):
#                 if k==0 or k==1:
#                     ndummy = Hex.LagrangeGaussLobatto(C,eps[k],z[i],z[j])[0]
#                     BasisBoundary[:,counter,k] = ndummy[:,0]

#                     dummy = Hex.GradLagrangeGaussLobatto(C,eps[k],z[i],z[j])
#                     gBasisBoundaryx[:,counter,k] = dummy[:,0]
#                     gBasisBoundaryy[:,counter,k] = dummy[:,1]
#                     gBasisBoundaryz[:,counter,k] = dummy[:,2]

#                 elif k==2 or k==3:
#                     ndummy = Hex.LagrangeGaussLobatto(C,z[i],eps[k],z[j])[0]
#                     BasisBoundary[:,counter,k] = ndummy[:,0]

#                     dummy = ThreeD.GradLagrangeGaussLobatto(C,z[i],eps[k],z[j])
#                     gBasisBoundaryx[:,counter,k] = dummy[:,0]
#                     gBasisBoundaryy[:,counter,k] = dummy[:,1]
#                     gBasisBoundaryz[:,counter,k] = dummy[:,2]

#                 elif k==4 or k==5:
#                     ndummy = Hex.LagrangeGaussLobatto(C,z[i],z[j],eps[k])[0]
#                     BasisBoundary[:,counter,k] = ndummy[:,0]

#                     dummy = Hex.GradLagrangeGaussLobatto(C,z[i],z[j],eps[k])
#                     gBasisBoundaryx[:,counter,k] = dummy[:,0]
#                     gBasisBoundaryy[:,counter,k] = dummy[:,1]
#                     gBasisBoundaryz[:,counter,k] = dummy[:,2]


#                 counter+=1

#     class Boundary(object):
#         """docstring for BasisBoundary"""
#         def __init__(self, arg):
#             super(BasisBoundary, self).__init__()
#             self.arg = arg
#         Basis  = BasisBoundary
#         gBasisx = gBasisBoundaryx
#         gBasisy = gBasisBoundaryy
#         gBasisz = gBasisBoundaryz


#     return Boundary




import numpy as np
from Florence.FunctionSpace.OneDimensional import Line as OneD
from Florence.MeshGeneration.NodeArrangement import NodeArrangementHex

def LagrangeGaussLobatto(C,zeta,eta,beta,arrange=1):
    """This routine computes stable higher order Lagrangian bases with Gauss-Lobatto-Legendre points
        Refer to: Spencer's Spectral hp elements for details"""

    Neta = np.zeros((C+2,1));   Nzeta = np.zeros((C+2,1)); Nbeta = np.zeros((C+2,1))

    Nzeta[:,0] = OneD.LagrangeGaussLobatto(C,zeta)[0]
    Neta[:,0] =  OneD.LagrangeGaussLobatto(C,eta)[0]
    Nbeta[:,0] =  OneD.LagrangeGaussLobatto(C,beta)[0]


    if arrange==1:
        # Bases = np.zeros(((C+2)**3,1))
        # # Arrange in counterclockwise
        # zeta_index, eta_index = GetCounterClockwiseIndices(C)
        # TBases = np.dot(Nzeta,Neta.T)
        # counter=0
        # for j in range(0,C+2):
        #     for i in range(0,(C+2)**2):
        #         Bases[counter] = Nbeta[j]*TBases[zeta_index[i],eta_index[i]]
        #         counter+=1

        node_arranger = NodeArrangementHex(C)[2]
        Bases = np.einsum('i,j,k', Nbeta[:,0], Neta[:,0], Nzeta[:,0]).flatten()
        Bases = Bases[node_arranger]
        Bases = Bases[:,None]

    elif arrange==0:
        Bases = np.zeros((C+2,C+2,C+2))
        for i in range(0,C+2):
            Bases[:,:,i] = Nbeta[i]*np.dot(Nzeta,Neta.T)
        Bases = Bases.reshape((C+2)**3,1)

    return Bases

def GradLagrangeGaussLobatto(C,zeta,eta,beta,arrange=1):
    """This routine computes stable higher order Lagrangian bases with Gauss-Lobatto-Legendre points
        Refer to: Spencer's Spectral hp elements for details"""

    gBases = np.zeros(((C+2)**3,3))
    Nzeta = np.zeros((C+2,1));  Neta = np.zeros((C+2,1));   Nbeta = np.zeros((C+2,1))
    gNzeta = np.zeros((C+2,1)); gNeta = np.zeros((C+2,1));  gNbeta = np.zeros((C+2,1))
    # Compute each from one-dimensional bases
    Nzeta[:,0] = OneD.LagrangeGaussLobatto(C,zeta)[0]
    Neta[:,0] = OneD.LagrangeGaussLobatto(C,eta)[0]
    Nbeta[:,0] = OneD.LagrangeGaussLobatto(C,beta)[0]
    gNzeta[:,0] = OneD.LagrangeGaussLobatto(C,zeta)[1]
    gNeta[:,0] = OneD.LagrangeGaussLobatto(C,eta)[1]
    gNbeta[:,0] = OneD.LagrangeGaussLobatto(C,beta)[1]

    # Ternsorial product
    if arrange==1:
        # # Arrange in counterclockwise
        # zeta_index, eta_index = GetCounterClockwiseIndices(C)
        # gBases1 = np.dot(gNzeta,Neta.T)
        # gBases2 = np.dot(Nzeta,gNeta.T)
        # gBases3 = np.dot(Nzeta,Neta.T)
        # counter=0
        # for j in range(0,C+2):
        #     for i in range(0,(C+2)**2):
        #         gBases[counter,0] = Nbeta[j]*gBases1[zeta_index[i],eta_index[i]]
        #         gBases[counter,1] = Nbeta[j]*gBases2[zeta_index[i],eta_index[i]]
        #         gBases[counter,2] = gNbeta[j]*gBases3[zeta_index[i],eta_index[i]]
                # counter+=1

        node_arranger = NodeArrangementHex(C)[2]
        # g0 = np.einsum('i,j,k', gNbeta[:,0], Neta[:,0], Nzeta[:,0]).flatten()
        # g1 = np.einsum('i,j,k', Nbeta[:,0], gNeta[:,0], Nzeta[:,0]).flatten()
        # g2 = np.einsum('i,j,k', Nbeta[:,0], Neta[:,0], gNzeta[:,0]).flatten()
        g0 = np.einsum('i,j,k', Nbeta[:,0], Neta[:,0], gNzeta[:,0]).flatten()
        g1 = np.einsum('i,j,k', Nbeta[:,0], gNeta[:,0], Nzeta[:,0]).flatten()
        g2 = np.einsum('i,j,k', gNbeta[:,0], Neta[:,0], Nzeta[:,0]).flatten()
        gBases[:,0] = g0[node_arranger]
        gBases[:,1] = g1[node_arranger]
        gBases[:,2] = g2[node_arranger]

    elif arrange==0:
        gBases1 = np.zeros((C+2,C+2,C+2)); gBases2 = np.zeros((C+2,C+2,C+2)); gBases3 = np.zeros((C+2,C+2,C+2))
        for i in range(0,C+2):
            gBases1[:,:,i] = Nbeta[i]*np.dot(gNzeta,Neta.T)
            gBases2[:,:,i] = Nbeta[i]*np.dot(Nzeta,gNeta.T)
            gBases3[:,:,i] = gNbeta[i]*np.dot(Nzeta,Neta.T)
        gBases1 = gBases1.reshape((C+2)**3,1)
        gBases2 = gBases2.reshape((C+2)**3,1)
        gBases3 = gBases3.reshape((C+2)**3,1)
        gBases[:,0]=gBases1
        gBases[:,1]=gBases2
        gBases[:,2]=gBases3


    return gBases


from __future__ import print_function, division
import numpy as np

def Sort_BEM(boundary_data,stiffness_K1,stiffness_K2):

    # SORTKERNEL MATRICES
    LHS2LHS=[]; LHS2RHS=[]; RHS2RHS=[]; RHS2LHS=[]
    # Count number of columns going back and forth
    for i in range(0,boundary_data.shape[0]):
        # Known variables going to RHS
        if boundary_data[i,0]!=-1:
            LHS2RHS=np.append(LHS2RHS,i)
        # Unknown variables remaining at LHS
        else:
            LHS2LHS = np.append(LHS2LHS,i)
        # Unknown variables going to LHS
        if boundary_data[i,1]==-1:
            RHS2LHS = np.append(RHS2LHS,i)
        # Known variables remaining at RHS
        else:
            RHS2RHS = np.append(RHS2RHS,i)

    LHS2RHS = np.copy(LHS2RHS).astype(np.int64)
    RHS2RHS = np.copy(RHS2RHS).astype(np.int64)


    # Make total LHS and RHS sizes and which columns comes from where
    total_LHS = np.zeros((LHS2LHS.shape[0]+RHS2LHS.shape[0],2),dtype=np.int64)
    total_RHS = np.zeros((LHS2RHS.shape[0]+RHS2RHS.shape[0],2),dtype=np.int64)

    # In second columns of total_LHS and total_RHS a 1 means in-place columns and a 0 means
    # columns coming from the other side

    # Sort LHS
    total_LHS[0:,0] = np.concatenate((LHS2LHS,RHS2LHS))
    for i in range(0,LHS2LHS.shape[0]):
        total_LHS[i,1] = 1
    sort_indices_LHS = np.argsort(total_LHS[0:,0])
    total_LHS[0:,0] = np.sort(total_LHS[0:,0])
    total_LHS[0:,1] = total_LHS[sort_indices_LHS,1]


    # Sort RHS
    total_RHS[0:,0] = np.concatenate((RHS2RHS,LHS2RHS))
    for i in range(0,RHS2RHS.shape[0]):
        total_RHS[i,1] = 1
    sort_indices_RHS = np.argsort(total_RHS[0:,0])
    total_RHS[0:,0] = np.sort(total_RHS[0:,0])
    total_RHS[0:,1] = total_RHS[sort_indices_RHS,1]

    # Now build global kernel matrices
    # Make two other kernel matrices knowing the dimensions from LHS2RHS, RHS2LHS and so on
    global_K1 = np.zeros((total_LHS.shape[0],total_LHS.shape[0]))
    global_K2 = np.zeros((total_RHS.shape[0],total_RHS.shape[0]))

    for i in range(0,total_LHS.shape[0]):
        if total_LHS[i,1]==1:
            global_K1[0:,i] = stiffness_K1[0:,total_LHS[i,0]]
        else:
            global_K1[0:,i] = -stiffness_K2[0:,total_LHS[i,0]]

    for i in range(0,total_RHS.shape[0]):
        if total_RHS[i,1]==1:
            global_K2[0:,i] = stiffness_K2[0:,total_RHS[i,0]]
        else:
            global_K2[0:,i] = -stiffness_K1[0:,total_RHS[i,0]]


    # Make the RHS vector (vector of known variables) to be multiplied with RHS matrix
    RHS_vector = np.zeros((total_RHS.shape[0],1))
    for i in range(0,total_RHS.shape[0]):
        if total_RHS[i,1]==1:
            RHS_vector[i,0] = boundary_data[total_RHS[i,0],1]
        else:
            RHS_vector[i,0] = boundary_data[total_RHS[i,0],0]

    mm=global_K2.dot(RHS_vector) # usual matrix-vector product (checked)

    return global_K1,mm, total_LHS,total_RHS, LHS2LHS, LHS2RHS, RHS2LHS, RHS2RHS